import { ConsoleMixin } from 'ferda/ConsoleMixin'
import { DatetimeMixin } from 'ferda/DatetimeMixin'
import { HttpMixin } from 'ferda/HttpMixin'
import { NotificationsMixin } from 'ferda/NotificationsMixin'
import { OBJECT_TYPE } from 'ferda/RegistryMixin'
import { URLS as FERDA_URLS } from 'ferda/Urls'
import { mapGetters } from 'vuex'

import { URLS } from 'ferda_verification/Urls'

export const VERIFICATION_STATUS = Object.freeze({
    CREATED: 'created',
    INITIALIZED: 'initialized',
    ACCESSED: 'accessed',
    PENDING: 'pending',
    CONSTRAINED: 'constrained',
    FUTILE: 'futile',
    SUCCEEDED: 'succeeded',
    FAILED: 'failed',
    CANCELED: 'canceled',
})

export const VERIFICATION_COLOR = Object.freeze({
    [VERIFICATION_STATUS.CREATED]: '#eeeeee',
    [VERIFICATION_STATUS.INITIALIZED]: '#eeeeee',
    [VERIFICATION_STATUS.ACCESSED]: '#98c6ea',
    [VERIFICATION_STATUS.PENDING]: '#98c6ea',
    [VERIFICATION_STATUS.CONSTRAINED]: '#fa93ab',
    [VERIFICATION_STATUS.FUTILE]: '#fa93ab',
    [VERIFICATION_STATUS.SUCCEEDED]: '#65e0b8',
    [VERIFICATION_STATUS.FAILED]: '#fa93ab',
    [VERIFICATION_STATUS.CANCELED]: '#98c6ea',
})

export const VERIFICATION_LIST_PARAMS = Object.freeze({ STATUSES: 'statuses' })

export const VERIFICATION_EVENT_TYPE = Object.freeze({
    FAILED: 'failed',
    FUTILE: 'futile',
    CANCELED: 'canceled',
    OUTZONE: 'outzone',
    VERIFY: 'verify_contact_data',
    USE_METHOD: 'use_method',
})

export const VERIFICATION_ACTION = Object.freeze({
    ADD: 'add_to_verifications',
    APPROVE_FUTILE: 'approve_futile',
    CANCEL: 'cancel',
    CONFIRM: 'confirm',
    DELAY_STATUS: 'delay_status',
    REJECT: 'reject',
    RESEND_EMAIL: 'resend_email',
})

export const VERIFICATION_MESSAGE_TYPE = Object.freeze({ VERIFICATION_INITIAL: 'verification_initial' })

export const VerificationMixin = {
    mixins: [ConsoleMixin, DatetimeMixin, HttpMixin, NotificationsMixin],
    computed: {
        ...mapGetters('general', ['permissions']),
    },
    created() {
        this.PAGE_NUMBERS = Object.freeze([10, 20, 50, -1])
        this.OBJECT_TYPE = OBJECT_TYPE
    },
    methods: {
        /**
         * Gets verification list
         *
         * @param { Object } params - URLSearchParams query
         * @returns { Object } - null in case of server error
         * @returns { Object } - response with data and pagination
         */
        async get_verification_list_data(params) {
            let response
            try {
                response = await this.http_get(URLS.verification_list_ajax(), { params })
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_BACKEND_NOT_AVAILABLE)
                return null
            }
            return response.data
        },
        /**
         * Gets verification action form title
         *
         * @param { String } last_action_type - last clicked action
         * @returns { String } - desired title
         */
        get_verification_form_title(last_action_type) {
            switch (last_action_type) {
            case VERIFICATION_ACTION.ADD:
                return this.$t('ferda_verification_common:add_to_verifications')
            case VERIFICATION_ACTION.APPROVE_FUTILE:
                return this.$t('ferda_verification_common:approve_futile')
            case VERIFICATION_ACTION.CANCEL:
                return this.$t('ferda_verification_common:cancel_verification')
            case VERIFICATION_ACTION.CONFIRM:
                return this.$t('ferda_verification_common:confirm_verification')
            case VERIFICATION_ACTION.DELAY_STATUS:
                return this.$t('ferda_verification_common:delay_status')
            case VERIFICATION_ACTION.REJECT:
                return this.$t('ferda_verification_common:reject_verification')
            case VERIFICATION_ACTION.RESEND_EMAIL:
                return this.$t('ferda_verification_common:resend_email')
            default:
                return ''
            }
        },
        /**
         * Gets verification list params
         *
         * @param { Array<String> } statuses - wanted statuses
         * @returns { Object } - URLSearchParams query
         */
        get_verification_list_params(statuses) {
            const params = new URLSearchParams()
            for (const status of statuses) {
                params.append(VERIFICATION_LIST_PARAMS.STATUSES, status)
            }
            return params
        },
        /**
         * Gets parameters for verification modification
         *
         * @param { String } last_action_type - desired action type
         * @param { Any } payload - any additional needed info
         * @returns { Object } - URLSearchParams query
         */
        get_verification_modify_params(last_action_type, payload = null) {
            const params = new URLSearchParams()
            switch (last_action_type) {
            case VERIFICATION_ACTION.RESEND_EMAIL:
                params.append('message_type', VERIFICATION_MESSAGE_TYPE.VERIFICATION_INITIAL)
                return params
            case VERIFICATION_ACTION.DELAY_STATUS:
                params.append('delay', payload)
                return params
            }
            return params
        },
        /**
         * Adds contact to verifications
         *
         * @param { String } contact_id - contact ID
         * @returns { Object } - null in case of server error
         */
        async add_to_verifications(contact_id) {
            try {
                await this.http_post(URLS.add_to_verifications_ajax(contact_id), {}, {
                    headers: { 'X-CSRFToken': this.$cookies.get('csrftoken') },
                })
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_BACKEND_NOT_AVAILABLE)
                return null
            }
        },
        /**
         * Approves domain deletion
         *
         * @param { String } verification_id - verification ID
         * @returns { Object } - null in case of server error
         */
        async approve_futile(verification_id) {
            try {
                await this.http_put(URLS.approve_futile_ajax(verification_id), {}, {
                    headers: { 'X-CSRFToken': this.$cookies.get('csrftoken') },
                })
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_BACKEND_NOT_AVAILABLE)
                return null
            }
        },
        /**
         * Cancels verification
         *
         * @param { String } verification_id - verification ID
         * @returns { Object } - null in case of server error
         */
        async cancel_verification(verification_id) {
            try {
                await this.http_put(URLS.cancel_verification_ajax(verification_id), {}, {
                    headers: { 'X-CSRFToken': this.$cookies.get('csrftoken') },
                })
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_BACKEND_NOT_AVAILABLE)
                return null
            }
        },
        /**
         * Gets verification by ID
         *
         * @param { String } verification_id - verification ID
         * @param { Object } abort_controller
         * @returns { Object } - null in case of server error
         * @returns { Object } - particular verification
         */
        async get_verification(verification_id, abort_controller = null) {
            let response
            try {
                response = await this.http_get(URLS.verification_detail_ajax(verification_id),
                    { signal: abort_controller?.signal || null })
            } catch (error) {
                const error_message = this.is_http_cancel(error) ?
                    this.INFO_REQUEST_ABORTED : this.ERROR_BACKEND_NOT_AVAILABLE
                this.log_error(error)
                this.notify_message(error_message)
                return null
            }
            return response.data
        },
        /**
         * Gets verification list by contact ID
         *
         * @param { String } contact_id - contact ID
         * @returns { Object } - null in case of server error
         * @returns { Object } - response with data and pagination
         */
        async get_verification_list_by_contact_id(contact_id) {
            let response
            try {
                response = await this.http_get(URLS.verification_list_ajax(), {
                    params: new URLSearchParams(`?contact_id=${contact_id}`),
                })
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_BACKEND_NOT_AVAILABLE)
                return null
            }
            return response.data
        },
        /**
         * Resends e-mail
         *
         * @param { String } verification_id - verification ID
         * @param { Object } params - URLSearchParams query
         * @returns { Object } - null in case of server error
         */
        async resend_email(verification_id, params) {
            try {
                await this.http_post(URLS.resend_email_ajax(verification_id), null, {
                    params,
                    headers: { 'X-CSRFToken': this.$cookies.get('csrftoken') },
                })
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_BACKEND_NOT_AVAILABLE)
                return null
            }
        },
        /**
         * Makes requests to delay status (both futile and constrained)
         *
         * @param { String } verification_id - verification ID
         * @param { Object } params - URLSearchParams query
         * @returns { Object } - null in case of server error
         */
        async delay_status(verification_id, params) {
            const params_futile = new URLSearchParams(params.toString())
            const params_constrained = new URLSearchParams(params.toString())
            params_futile.append('status', VERIFICATION_STATUS.FUTILE)
            params_constrained.append('status', VERIFICATION_STATUS.CONSTRAINED)

            try {
                await this.http_put(URLS.delay_status_ajax(verification_id), null, {
                    params: params_futile,
                    headers: { 'X-CSRFToken': this.$cookies.get('csrftoken') },
                })
                await this.http_put(URLS.delay_status_ajax(verification_id), null, {
                    params: params_constrained,
                    headers: { 'X-CSRFToken': this.$cookies.get('csrftoken') },
                })
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_BACKEND_NOT_AVAILABLE)
                return null
            }
        },
        /**
         * Makes particular action above verification
         *
         * @param { String } last_action_type - last action type
         * @param { Object } params - URLSearchParams query
         * @param { String } verification_id - verification ID
         * @param { String } contact_id - contact ID
         * @param { Array } confirm_data - data form manual check verification
         * @returns { Object } - info after success operation
         * @returns { Object } - null in case of server error or unknown action type
         */
        async make_verification_action(
            last_action_type,
            params,
            verification_id,
            contact_id = null,
            confirm_data = null,
        ) {
            switch (last_action_type) {
            case VERIFICATION_ACTION.ADD:
                return await this.add_to_verifications(contact_id)
            case VERIFICATION_ACTION.APPROVE_FUTILE:
                return await this.approve_futile(verification_id)
            case VERIFICATION_ACTION.CANCEL:
                return await this.cancel_verification(verification_id)
            case VERIFICATION_ACTION.CONFIRM:
                return await this.manual_check_verification(true, verification_id, confirm_data)
            case VERIFICATION_ACTION.DELAY_STATUS:
                return await this.delay_status(verification_id, params)
            case VERIFICATION_ACTION.REJECT:
                return await this.manual_check_verification(false, verification_id, confirm_data)
            case VERIFICATION_ACTION.RESEND_EMAIL:
                return await this.resend_email(verification_id, params)
            default:
                return null
            }
        },
        /**
         * Gets specified date for particular verification
         *
         * @param { Object } - object with specified events and next events
         * @param { String } desired_event - event to be searched
         * @param { Boolean } multiple - whether there can be more events
         * @returns { Array<Object> } - in case there are more events
         * @returns { Object } - null in case of missing event
         * @returns { Object } - particular event
         */
        get_verification_event({ events, next_events }, desired_event, multiple = false) {
            if (multiple) {
                // In case of multipe possible events, list all of them from events and next_events
                return [...events, ...next_events].filter(event => event.type === desired_event)
            }

            // In case of single event, find last in next_events or last in events
            const event = next_events.findLast(event => event.type === desired_event) ||
                events.findLast(event => event.type === desired_event)
            return event || null
        },
        /**
         * Checks whether user can cancel verification
         *
         * @param { String } status - verification status
         * @returns { Boolean } - can or can't cancel
         */
        can_cancel(status) {
            return status !== VERIFICATION_STATUS.CANCELED && this.permissions.includes('ferda_verification.can_cancel')
        },
        /**
         * Checks whether user can add contact to verifications
         * Allowed are only contacts not already included in verifications and not deleted
         *
         * @param { Object } contact - info about contact itself
         * @returns { Boolean } - can or can't create
         */
        can_create(contact) {
            const has_permission = this.permissions.includes('ferda_verification.can_create')
            const is_deleted = !!contact.object_info.events.unregistered?.timestamp

            return has_permission && !is_deleted
        },
        /**
         * Checks whether user can resend e-mail
         *
         * @param { String } status - verification status
         * @returns { Boolean } - can or can't resend
         */
        can_resend(status) {
            return [
                VERIFICATION_STATUS.ACCESSED,
                VERIFICATION_STATUS.CONSTRAINED,
                VERIFICATION_STATUS.FUTILE,
                VERIFICATION_STATUS.INITIALIZED,
                VERIFICATION_STATUS.PENDING,
            ].includes(status) && this.permissions.includes('ferda_verification.can_resend_message')
        },
        /**
         * Checks whether user can delay status
         *
         * @returns { Boolean } - can or can't delay
         */
        can_delay() {
            return this.permissions.includes('ferda_verification.can_delay_status')
        },
        can_approve_futile() {
            return this.permissions.includes('ferda_verification.can_approve_futile')
        },
        /**
         * Checks whether user can resolve manual check verification
         *
         * @returns { Bolean } - can or can't resolve manual check
         */
        can_resolve() {
            return this.permissions.includes('ferda_verification.can_resolve_manual_check')
        },
        /**
         * Sorts table columns
         *
         * @param { Array<Object> } items - table data
         * @param { Array<String> } sort_by - array of columns table is sorted by (always just 1)
         * @param { Array<Boolean> } is_desc - array of columns direction sort (always just 1)
         * @returns { Array<Object> } - sorted table data
         */
        sort_table(items, sort_by, is_desc) {
            return items.sort((first, second) => {
                if (sort_by[0] === 'create_datetime') {
                    const first_date = new Date(first[sort_by])
                    const second_date = new Date(second[sort_by])
                    return is_desc[0] ? second_date - first_date : first_date - second_date
                }

                if (sort_by[0] === 'events') {
                    const first_date = new Date(this.get_verification_event(
                        first, VERIFICATION_EVENT_TYPE.OUTZONE,
                    ).event_datetime)

                    const second_date = new Date(this.get_verification_event(
                        second, VERIFICATION_EVENT_TYPE.OUTZONE,
                    ).event_datetime)
                    return is_desc[0] ?  second_date - first_date : first_date - second_date
                }

                if (sort_by[0] === 'manual_check_request') {
                    return is_desc[0] ? first[sort_by] - second[sort_by] : second[sort_by] - first[sort_by]
                }

                if (typeof first[sort_by] !== 'undefined') {
                    const first_val = first[sort_by].toLowerCase().trim()
                    const second_val = second[sort_by].toLowerCase().trim()
                    return is_desc[0] ? second_val.localeCompare(first_val) : first_val.localeCompare(second_val)
                }
                return items
            })
        },
        /**
         * Gets additional contact info
         *
         * @returns { Object } - null in case of server error
         * @returns { Object } - detail of contact
         */
        async get_contact_detail(contact_id) {
            let response
            try {
                response = await this.http_get(FERDA_URLS.object_info_ajax(OBJECT_TYPE.CONTACT, contact_id))
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_BACKEND_NOT_AVAILABLE)
                return null
            }
            return response.data
        },
        /**
         * Manually confirm or reject verification
         *
         * @param { Boolean } verified - confirmation or rejection of data
         * @param { String } verification_id
         * @param { Object } confirm_data
         * @returns { Object } - null in case of server error
        */
        async manual_check_verification(verified, verification_id, confirm_data) {
            const data = {
                verified: verified,
                contact_data: confirm_data.data,
                contact_history_id: confirm_data.contact_history_id,
            }
            try {
                await this.http_put(URLS.confirm_verification_ajax(verification_id),
                    data, { headers: { 'X-CSRFToken': this.$cookies.get('csrftoken') }},
                )
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_BACKEND_NOT_AVAILABLE)
                return null
            }
        },
    },
}
