import json
from datetime import date, datetime, timezone
from unittest.mock import call, patch
from urllib.parse import urlencode

from django.contrib.auth import get_user_model
from django.test import Client, TestCase, override_settings, tag
from django.urls import reverse
from ferda.api import PydanticJsonEncoder
from ferda.constants import LoggerResult
from ferda.tests.utils import LoggerMixin
from fred_types import ContactHistoryId, ContactId
from grill import LogProperty
from regal import Address, Contact
from verifier import (
    ContactData,
    PaginatedList,
    Pagination,
    Verification,
    VerificationId,
    VerificationResult,
    VerificationStatus,
)
from verifier.exceptions import (
    ContactDoesNotExist,
    IncorrectStatus,
    InvalidData,
    VerificationDoesNotExist,
)

from ferda_verification.api import VerificationWithContactHandle
from ferda_verification.constants import LogEntryType, LogReferenceType

User = get_user_model()


def get_verification(**kwargs):
    defaults = {
        "verification_id": "v:1",
        "contact_id": "c:1",
        "create_datetime": datetime(2022, 1, 1, tzinfo=timezone.utc),
    }
    defaults.update(**kwargs)
    return Verification(**defaults)


def get_verification_result(**kwargs):
    data_kwargs = kwargs.pop("data", {})
    return VerificationResult(
        verify_datetime=datetime(2022, 1, 1, tzinfo=timezone.utc),
        method="letter",
        data=ContactData(**data_kwargs),
        contact_history_id="ch:1",
        **kwargs,
    )


@tag("api")
@override_settings(ROOT_URLCONF="ferda_verification.tests.urls")
class ListTest(LoggerMixin, TestCase):
    def setUp(self):
        self.client = Client(json_encoder=PydanticJsonEncoder)

        super().setUp()
        user = User.objects.create_superuser("admin")
        self.client.force_login(user)

    def _test_list(self, verification_client, contact_client, query, pagination):
        verification = get_verification()
        verification_client.list.return_value = PaginatedList[Verification](
            results=[verification], pagination=pagination
        )
        contact_client.get_contact_info.return_value = Contact(
            contact_id=ContactId("c:1"),
            contact_handle="RIMMER",
        )

        response = self.client.get(
            reverse("ferda_verification:api:list") + "?" + urlencode(query, doseq=True)
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "results": [
                    json.loads(
                        VerificationWithContactHandle(
                            **verification.dict(), contact_handle="RIMMER"
                        ).json()
                    )
                ],
                "pagination": pagination,
            },
        )
        self.assertEqual(
            verification_client.mock_calls,
            [
                call.list(
                    **{
                        "contact_id": None,
                        "statuses": None,
                        "manual_check_request": None,
                        "page_size": None,
                        "page_token": None,
                        **query,
                    }
                )
            ],
        )

    @patch("ferda_verification.utils.CONTACT", autospec=True)
    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_list(self, verification_client, contact_client):
        self._test_list(verification_client, contact_client, query={}, pagination=None)
        self.assertLog(
            LogEntryType.VERIFICATION_LIST,
            LoggerResult.SUCCESS,
            input_properties={},
            output_references={
                LogReferenceType.VERIFICATION: [get_verification().verification_id],
            },
        )

    @patch("ferda_verification.utils.CONTACT", autospec=True)
    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_list_full(self, verification_client, contact_client):
        self._test_list(
            verification_client,
            contact_client,
            query={
                "contact_id": "c:1",
                "statuses": [
                    VerificationStatus.INITIALIZED.value,
                    VerificationStatus.ACCESSED.value,
                ],
                "manual_check_request": True,
                "page_size": 10,
                "page_token": "Starbug",
            },
            pagination=Pagination(next_page_token="Starbug2", items_left=1),
        )
        self.assertLog(
            LogEntryType.VERIFICATION_LIST,
            LoggerResult.SUCCESS,
            input_properties={
                "contact_id": "c:1",
                "statuses": [
                    VerificationStatus.INITIALIZED.value,
                    VerificationStatus.ACCESSED.value,
                ],
                "manual_check_request": str(True),
                "page_size": "10",
                "page_token": "Starbug",
            },
            output_references={
                LogReferenceType.VERIFICATION: [get_verification().verification_id],
            },
        )


@tag("api")
@override_settings(ROOT_URLCONF="ferda_verification.tests.urls")
class CreateTest(LoggerMixin, TestCase):
    def setUp(self):
        super().setUp()
        user = User.objects.create_superuser("admin")
        self.client.force_login(user)

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_create(self, verification_client):
        verification = get_verification(contact_id="c:1")
        verification_client.create.return_value = verification

        response = self.client.post(
            reverse("ferda_verification:api:create") + "?" + urlencode({"contact_id": "c:1"})
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(Verification(**response.json()), verification)
        self.assertEqual(
            verification_client.mock_calls, [call.create(contact_id=ContactId("c:1"))]
        )
        self.assertLog(
            LogEntryType.VERIFICATION_CREATE,
            LoggerResult.SUCCESS,
            input_properties={"contact_id": "c:1"},
            output_references={
                LogReferenceType.VERIFICATION: verification.verification_id,
                LogReferenceType.CONTACT: "c:1",
            },
        )

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_create_contact_does_not_exist(self, verification_client):
        verification_client.create.side_effect = ContactDoesNotExist

        response = self.client.post(
            reverse("ferda_verification:api:create") + "?" + urlencode({"contact_id": "c:1"})
        )

        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.json(), {"detail": "Contact not found"})
        self.assertEqual(
            verification_client.mock_calls, [call.create(contact_id=ContactId("c:1"))]
        )
        self.assertLog(
            LogEntryType.VERIFICATION_CREATE,
            LoggerResult.FAIL,
            input_properties={"contact_id": "c:1"},
        )


@tag("api")
@override_settings(ROOT_URLCONF="ferda_verification.tests.urls")
class GetTest(LoggerMixin, TestCase):
    def setUp(self):
        super().setUp()
        user = User.objects.create_superuser("admin")
        self.client.force_login(user)

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_get(self, verification_client):
        verification = get_verification()
        verification_client.get.return_value = verification

        response = self.client.get(
            reverse(
                "ferda_verification:api:get",
                kwargs={"verification_id": verification.verification_id},
            )
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(Verification(**response.json()), verification)
        self.assertEqual(verification_client.mock_calls, [call.get(verification.verification_id)])
        self.assertLog(
            LogEntryType.VERIFICATION_GET,
            LoggerResult.SUCCESS,
            input_properties={"verification_id": verification.verification_id},
            output_references={
                LogReferenceType.VERIFICATION: verification.verification_id,
                LogReferenceType.CONTACT: verification.contact_id,
            },
        )

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_get_does_not_exist(self, verification_client):
        verification_client.get.side_effect = VerificationDoesNotExist

        response = self.client.get(
            reverse("ferda_verification:api:get", kwargs={"verification_id": "v:1"})
        )

        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.json(), {"detail": "Verification not found"})
        self.assertEqual(verification_client.mock_calls, [call.get(VerificationId("v:1"))])
        self.assertLog(
            LogEntryType.VERIFICATION_GET,
            LoggerResult.FAIL,
            input_properties={"verification_id": "v:1"},
        )


@tag("api")
@override_settings(ROOT_URLCONF="ferda_verification.tests.urls")
class CancelTest(LoggerMixin, TestCase):
    def setUp(self):
        super().setUp()
        user = User.objects.create_superuser("admin")
        self.client.force_login(user)

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_cancel(self, verification_client):
        verification_client.cancel.return_value = None

        response = self.client.put(
            reverse(
                "ferda_verification:api:cancel",
                kwargs={"verification_id": "v:1"},
            )
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"")
        self.assertEqual(verification_client.mock_calls, [call.cancel("v:1")])
        self.assertLog(
            LogEntryType.VERIFICATION_CANCEL,
            LoggerResult.SUCCESS,
            input_properties={"verification_id": "v:1"},
            output_references={
                LogReferenceType.VERIFICATION: "v:1",
            },
        )

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_cancel_does_not_exist(self, verification_client):
        verification_client.cancel.side_effect = VerificationDoesNotExist

        response = self.client.put(
            reverse(
                "ferda_verification:api:cancel",
                kwargs={"verification_id": "v:1"},
            )
        )

        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.json(), {"detail": "Verification not found"})
        self.assertEqual(verification_client.mock_calls, [call.cancel(VerificationId("v:1"))])
        self.assertLog(
            LogEntryType.VERIFICATION_CANCEL,
            LoggerResult.FAIL,
            input_properties={"verification_id": "v:1"},
        )

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_cancel_incorrect_status(self, verification_client):
        verification_client.cancel.side_effect = IncorrectStatus

        response = self.client.put(
            reverse(
                "ferda_verification:api:cancel",
                kwargs={"verification_id": "v:1"},
            )
        )

        self.assertEqual(response.status_code, 409)
        self.assertEqual(response.json(), {"detail": "Incorrect verification status"})
        self.assertEqual(verification_client.mock_calls, [call.cancel(VerificationId("v:1"))])
        self.assertLog(
            LogEntryType.VERIFICATION_CANCEL,
            LoggerResult.FAIL,
            input_properties={"verification_id": "v:1"},
        )


@tag("api")
@override_settings(ROOT_URLCONF="ferda_verification.tests.urls")
class ApproveFutileTest(LoggerMixin, TestCase):
    def setUp(self):
        super().setUp()
        user = User.objects.create_superuser("admin")
        self.client.force_login(user)

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_approve_futile(self, verification_client):
        verification_client.approve_futile.return_value = None

        response = self.client.put(
            reverse(
                "ferda_verification:api:approve_futile",
                kwargs={"verification_id": "v:1"},
            )
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"")
        self.assertEqual(verification_client.mock_calls, [call.approve_futile("v:1")])
        self.assertLog(
            LogEntryType.VERIFICATION_APPROVE_FUTILE,
            LoggerResult.SUCCESS,
            input_properties={"verification_id": "v:1"},
            output_references={
                LogReferenceType.VERIFICATION: "v:1",
            },
        )

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_approve_futile_does_not_exist(self, verification_client):
        verification_client.approve_futile.side_effect = VerificationDoesNotExist

        response = self.client.put(
            reverse(
                "ferda_verification:api:approve_futile",
                kwargs={"verification_id": "v:1"},
            )
        )

        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.json(), {"detail": "Verification not found"})
        self.assertEqual(
            verification_client.mock_calls, [call.approve_futile(VerificationId("v:1"))]
        )
        self.assertLog(
            LogEntryType.VERIFICATION_APPROVE_FUTILE,
            LoggerResult.FAIL,
            input_properties={"verification_id": "v:1"},
        )

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_approve_futile_incorrect_status(self, verification_client):
        verification_client.approve_futile.side_effect = IncorrectStatus

        response = self.client.put(
            reverse(
                "ferda_verification:api:approve_futile",
                kwargs={"verification_id": "v:1"},
            )
        )

        self.assertEqual(response.status_code, 409)
        self.assertEqual(response.json(), {"detail": "Incorrect verification status"})
        self.assertEqual(
            verification_client.mock_calls, [call.approve_futile(VerificationId("v:1"))]
        )
        self.assertLog(
            LogEntryType.VERIFICATION_APPROVE_FUTILE,
            LoggerResult.FAIL,
            input_properties={"verification_id": "v:1"},
        )


@tag("api")
@override_settings(ROOT_URLCONF="ferda_verification.tests.urls")
class DelayStatusTest(LoggerMixin, TestCase):
    def setUp(self):
        super().setUp()
        user = User.objects.create_superuser("admin")
        self.client.force_login(user)

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_delay_status(self, verification_client):
        verification_client.delay_status.return_value = None

        response = self.client.put(
            reverse(
                "ferda_verification:api:delay_status",
                kwargs={"verification_id": "v:1"},
            )
            + "?"
            + urlencode({"status": VerificationStatus.CONSTRAINED.value, "delay": 300})
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"")
        self.assertEqual(
            verification_client.mock_calls,
            [
                call.delay_status(
                    VerificationId("v:1"), status=VerificationStatus.CONSTRAINED, delay=300
                )
            ],
        )
        self.assertLog(
            LogEntryType.VERIFICATION_DELAY_STATUS,
            LoggerResult.SUCCESS,
            input_properties={
                "verification_id": "v:1",
                "status": VerificationStatus.CONSTRAINED.value,
                "delay": "300",
            },
            output_references={
                LogReferenceType.VERIFICATION: "v:1",
            },
        )

    def _test_error(self, verification_client, error, status_code, json):
        verification_client.delay_status.side_effect = error

        response = self.client.put(
            reverse(
                "ferda_verification:api:delay_status",
                kwargs={"verification_id": "v:1"},
            )
            + "?"
            + urlencode({"status": VerificationStatus.CONSTRAINED.value, "delay": 300})
        )

        self.assertEqual(response.status_code, status_code)
        self.assertEqual(response.json(), json)
        self.assertEqual(
            verification_client.mock_calls,
            [
                call.delay_status(
                    VerificationId("v:1"), status=VerificationStatus.CONSTRAINED, delay=300
                )
            ],
        )
        self.assertLog(
            LogEntryType.VERIFICATION_DELAY_STATUS,
            LoggerResult.FAIL,
            input_properties={
                "verification_id": "v:1",
                "status": VerificationStatus.CONSTRAINED.value,
                "delay": "300",
            },
        )

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_delay_status_does_not_exist(self, verification_client):
        self._test_error(
            verification_client,
            VerificationDoesNotExist,
            404,
            {"detail": "Verification not found"},
        )

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_delay_status_incorrect_status(self, verification_client):
        self._test_error(
            verification_client, IncorrectStatus, 409, {"detail": "Incorrect verification status"}
        )


@tag("api")
@override_settings(ROOT_URLCONF="ferda_verification.tests.urls")
class ListVerificationResultsTest(LoggerMixin, TestCase):
    def setUp(self):
        super().setUp()
        user = User.objects.create_superuser("admin")
        self.client.force_login(user)

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_get(self, verification_client):
        result = get_verification_result()
        verification_client.list_verification_results.return_value = [result]

        response = self.client.get(
            reverse(
                "ferda_verification:api:list_verification_results",
                kwargs={"verification_id": "v:1"},
            )
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual([VerificationResult(**result) for result in response.json()], [result])
        self.assertEqual(verification_client.mock_calls, [call.list_verification_results("v:1")])
        self.assertLog(
            LogEntryType.VERIFICATION_LIST_RESULTS,
            LoggerResult.SUCCESS,
            input_properties={"verification_id": "v:1"},
            output_references={
                LogReferenceType.VERIFICATION: "v:1",
            },
        )

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_get_does_not_exist(self, verification_client):
        verification_client.list_verification_results.side_effect = VerificationDoesNotExist

        response = self.client.get(
            reverse(
                "ferda_verification:api:list_verification_results",
                kwargs={"verification_id": "v:1"},
            )
        )

        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.json(), {"detail": "Verification not found"})
        self.assertEqual(
            verification_client.mock_calls, [call.list_verification_results(VerificationId("v:1"))]
        )
        self.assertLog(
            LogEntryType.VERIFICATION_LIST_RESULTS,
            LoggerResult.FAIL,
            input_properties={"verification_id": "v:1"},
        )


@tag("api")
@override_settings(ROOT_URLCONF="ferda_verification.tests.urls")
class ResolveManualCheckTest(LoggerMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.client = Client(json_encoder=PydanticJsonEncoder)

        user = User.objects.create_superuser("admin")
        self.client.force_login(user)

        self.person_data = {
            "verified": True,
            "contact_data": {
                "name": "Arnold Rimmer",
                "birthdate": "2022-01-01",
                "emails": ["rimmer@example.org", "arnold@example.org"],
            },
            "contact_history_id": "ch:1",
        }

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_person(self, verification_client):
        verification_client.resolve_manual_check.return_value = None

        response = self.client.put(
            reverse(
                "ferda_verification:api:resolve_manual_check",
                kwargs={"verification_id": "v:1"},
            ),
            data=self.person_data,
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"")
        self.assertEqual(
            verification_client.mock_calls,
            [
                call.resolve_manual_check(
                    VerificationId("v:1"),
                    verified=True,
                    contact_data=ContactData(
                        name="Arnold Rimmer",
                        birthdate=date(2022, 1, 1),
                        emails=["rimmer@example.org", "arnold@example.org"],
                    ),
                    contact_history_id=ContactHistoryId("ch:1"),
                )
            ],
        )
        self.assertLog(
            LogEntryType.VERIFICATION_RESOLVE_MANUAL_CHECK,
            LoggerResult.SUCCESS,
            input_properties={
                "verification_id": "v:1",
                "verified": str(True),
                "name": "Arnold Rimmer",
                "birthdate": "2022-01-01",
                "emails": "rimmer@example.org, arnold@example.org",
                "contact_history_id": "ch:1",
            },
            output_references={
                LogReferenceType.VERIFICATION: "v:1",
            },
        )

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_company(self, verification_client):
        verification_client.resolve_manual_check.return_value = None

        company_data = {
            "verified": False,
            "contact_data": {
                "organization": "Jupiter Mining Corporation",
                "company_registration_number": "JMC-01",
                "address": Address(street=["Sector A", "Deck 4"], city="Red Dwarf"),
            },
            "contact_history_id": "ch:1",
        }

        response = self.client.put(
            reverse(
                "ferda_verification:api:resolve_manual_check",
                kwargs={"verification_id": "v:1"},
            ),
            data=company_data,
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"")
        self.assertEqual(
            verification_client.mock_calls,
            [
                call.resolve_manual_check(
                    VerificationId("v:1"),
                    verified=False,
                    contact_data=ContactData(
                        organization="Jupiter Mining Corporation",
                        company_registration_number="JMC-01",
                        address=Address(street=["Sector A", "Deck 4"], city="Red Dwarf"),
                    ),
                    contact_history_id=ContactHistoryId("ch:1"),
                )
            ],
        )
        self.assertLog(
            LogEntryType.VERIFICATION_RESOLVE_MANUAL_CHECK,
            LoggerResult.SUCCESS,
            input_properties={
                "verification_id": "v:1",
                "verified": str(False),
                "organization": "Jupiter Mining Corporation",
                "company_registration_number": "JMC-01",
                "address": LogProperty(
                    "",
                    children={
                        "company": "",
                        "street": "Sector A, Deck 4",
                        "city": "Red Dwarf",
                        "state_or_province": "",
                        "postal_code": "",
                        "country_code": "",
                    },
                ),
                "contact_history_id": "ch:1",
            },
            output_references={
                LogReferenceType.VERIFICATION: "v:1",
            },
        )

    def _test_error(self, verification_client, error, status_code, json):
        verification_client.resolve_manual_check.side_effect = error

        response = self.client.put(
            reverse(
                "ferda_verification:api:resolve_manual_check",
                kwargs={"verification_id": "v:1"},
            ),
            data=self.person_data,
            content_type="application/json",
            json_encoder=PydanticJsonEncoder,
        )

        self.assertEqual(response.status_code, status_code)
        self.assertEqual(response.json(), json)
        self.assertEqual(
            verification_client.mock_calls,
            [
                call.resolve_manual_check(
                    VerificationId("v:1"),
                    verified=True,
                    contact_data=ContactData(
                        name="Arnold Rimmer",
                        birthdate=date(2022, 1, 1),
                        emails=["rimmer@example.org", "arnold@example.org"],
                    ),
                    contact_history_id=ContactHistoryId("ch:1"),
                )
            ],
        )
        self.assertLog(
            LogEntryType.VERIFICATION_RESOLVE_MANUAL_CHECK,
            LoggerResult.FAIL,
            input_properties={
                "verification_id": "v:1",
                "verified": str(True),
                "name": "Arnold Rimmer",
                "birthdate": "2022-01-01",
                "emails": "rimmer@example.org, arnold@example.org",
                "contact_history_id": "ch:1",
            },
        )

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_invalid_data(self, verification_client):
        self._test_error(
            verification_client,
            InvalidData(fields=["contact_data.birthdate"]),
            400,
            {"detail": "Invalid data", "fields": ["contact_data.birthdate"]},
        )

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_verification_does_not_exist(self, verification_client):
        self._test_error(
            verification_client,
            VerificationDoesNotExist,
            404,
            {"detail": "Verification not found"},
        )

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_incorrect_status(self, verification_client):
        self._test_error(
            verification_client,
            IncorrectStatus,
            409,
            {"detail": "Incorrect verification status"},
        )


@tag("api")
@override_settings(ROOT_URLCONF="ferda_verification.tests.urls")
class ResendMessageTest(LoggerMixin, TestCase):
    def setUp(self):
        super().setUp()
        user = User.objects.create_superuser("admin")
        self.client.force_login(user)

    def _test_resend_message(self, verification_client, status_code):
        response = self.client.post(
            reverse(
                "ferda_verification:api:resend_message",
                kwargs={"verification_id": "v:1"},
            )
            + "?"
            + urlencode({"message_type": "initial"}),
        )

        self.assertEqual(response.status_code, status_code)
        self.assertEqual(
            verification_client.mock_calls,
            [call.resend_message(VerificationId("v:1"), message_type="initial")],
        )

        return response

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_resend_message_success(self, verification_client):
        verification_client.resend_message.return_value = None

        response = self._test_resend_message(verification_client, 200)

        self.assertEqual(response.content, b"")
        self.assertLog(
            LogEntryType.VERIFICATION_RESEND_MESSAGE,
            LoggerResult.SUCCESS,
            input_properties={"verification_id": "v:1", "message_type": "initial"},
            output_references={LogReferenceType.VERIFICATION: "v:1"},
        )

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_resend_message_invalid_data(self, verification_client):
        verification_client.resend_message.side_effect = InvalidData(fields=["message_type"])

        response = self._test_resend_message(verification_client, 400)

        self.assertEqual(response.json(), {"detail": "Invalid data", "fields": ["message_type"]})
        self.assertLog(
            LogEntryType.VERIFICATION_RESEND_MESSAGE,
            LoggerResult.FAIL,
            input_properties={"verification_id": "v:1", "message_type": "initial"},
            output_references={LogReferenceType.VERIFICATION: "v:1"},
        )

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_resend_message_does_not_exist(self, verification_client):
        verification_client.resend_message.side_effect = VerificationDoesNotExist

        response = self._test_resend_message(verification_client, 404)

        self.assertEqual(response.json(), {"detail": "Verification not found"})
        self.assertLog(
            LogEntryType.VERIFICATION_RESEND_MESSAGE,
            LoggerResult.FAIL,
            input_properties={"verification_id": "v:1", "message_type": "initial"},
        )

    @patch("ferda_verification.api.VERIFICATION", autospec=True)
    def test_resend_message_incorrect_status(self, verification_client):
        verification_client.resend_message.side_effect = IncorrectStatus

        response = self._test_resend_message(verification_client, 409)

        self.assertEqual(response.json(), {"detail": "Incorrect verification status"})
        self.assertLog(
            LogEntryType.VERIFICATION_RESEND_MESSAGE,
            LoggerResult.FAIL,
            input_properties={"verification_id": "v:1", "message_type": "initial"},
            output_references={LogReferenceType.VERIFICATION: "v:1"},
        )
